#!/bin/bash
# Optimise kernel parameters for running SAP HANA and HANA based products (such as Business One).
# The calculations are based on:
# - Parameters tuned by SAP installation wizard and configure_HANA.sh.
# - Various SAP notes.
# For SAP Netweaver tuning, please use "sap-netweaver" profile instead of this one.
# Authors: Howard Guo <hguo@suse.com>

. /usr/lib/tuned/functions

start() {
    # Read system memory size in MB
    declare -ri MEMSIZE=$(($(grep MemTotal /proc/meminfo | awk '{print $2}')/1024/1024))
    # Determine an appropriate value for kernel.shmmni
    declare -i SHMMNI=$(sysctl -n kernel.shmmni)
    if [ "$MEMSIZE" -lt 64 ]; then
        declare -i SHMMNI_NEW=4096
    elif [ "$MEMSIZE" -lt 256 ]; then
        declare -i SHMMNI_NEW=65536
    else
        declare -i SHMMNI_NEW=524288
    fi
    # New SHMMNI may not be lower than current settings
    if [ "$SHMMNI_NEW" -lt "$SHMMNI" ]; then
        declare -i SHMMNI_NEW="$SHMMNI"
    fi

    # SAP note 1557506 - Linux paging improvements
    # Tune pagecache_limit_mb and set pagecache_limit_ignore_dirty = 2
    # Recommended page cache limit = 2% of system memory
    declare -ri PAGECACHE_LIMIT=$(sysctl -n vm.pagecache_limit_mb)
    declare -ri PAGECACHE_LIMIT_NEW=$((${MEMSIZE}*1024*2/100))
    if [ "$PAGECACHE_LIMIT_NEW" -lt "$PAGECACHE_LIMIT" ]; then
        declare -i PAGECACHE_LIMIT_NEW="$PAGECACHE_LIMIT"
    fi

    # Apply new values
    save_value kernel.shmmni "$SHMMNI"
    sysctl -w "kernel.shmmni=$SHMMNI_NEW"
    save_value vm.pagecache_limit_mb "$PAGECACHE_LIMIT"
    sysctl -w "vm.pagecache_limit_mb=$PAGECACHE_LIMIT_NEW"
    save_value vm.pagecache_limit_ignore_dirty $(sysctl -n vm.pagecache_limit_ignore_dirty)
    sysctl -w "vm.pagecache_limit_ignore_dirty=2"

    # SAP note 2205917 - KSM and AutoNUMA both should be off
    save_value ksm $(cat /sys/kernel/mm/ksm/run)
    echo 0 > /sys/kernel/mm/ksm/run
    save_value numa_balancing $(cat /proc/sys/kernel/numa_balancing)
    echo 0 > /proc/sys/kernel/numa_balancing

    return 0
}

stop() {
    SHMMNI=$(restore_value kernel.shmmni)
    [ "$SHMMNI" ] && sysctl -w "kernel.shmmni=$SHMMNI"

    PAGECACHE_LIMIT=$(restore_value vm.pagecache_limit_mb)
    [ "$PAGECACHE_LIMIT" ] && sysctl -w "vm.pagecache_limit_mb=$PAGECACHE_LIMIT"
    PAGECACHE_LIMIT_IGNORE_DIRTY=$(restore_value vm.pagecache_limit_ignore_dirty)
    [ "$PAGECACHE_LIMIT_IGNORE_DIRTY" ] && sysctl -w "vm.pagecache_limit_ignore_dirty=$PAGECACHE_LIMIT_IGNORE_DIRTY"

    # Restore KSM and AutoNUMA settings
    KSM=$(restore_value ksm)
    [ "$KSM" ] && echo "$KSM" > /sys/kernel/mm/ksm/run
    NUMA_BALANCING=$(restore_value numa_balancing)
    [ "$NUMA_BALANCING" ] && echo "$NUMA_BALANCING" > /proc/sys/kernel/numa_balancing

    return 0
}

process $@
