#!/bin/bash
# Optimise kernel parameters, tmpfs, and system limits for running SAP Netweaver products.
# The calculations are based on:
# - Parameters tuned by the traditional "sapconf" package.
# - Various SAP notes.
# For SAP HANA tuning, please use "sap-hana" profile instead of this one.
# Authors: Howard Guo <hguo@suse.com>

. /usr/lib/tuned/functions

start() {
    # Read total memory size (including swap) in KBytes
    declare -ri VSZ=$(awk -v t=0 '/^(Mem|Swap)Total:/ {t+=$2} END {print t}' < /proc/meminfo)
    declare -ri PSZ=$(getconf PAGESIZE)

    # These names are the tuning parameters
    declare -r TUNED_VARS="TMPFS_SIZE SHMALL SEMMSL SEMMNS SEMOPM SEMMNI SHMMAX MAX_MAP_COUNT"
    for name in $TUNED_VARS; do
        # Current value
        declare -i $name=0
        # Calculated recommended value / future value
        declare -i ${name}_REQ=0
        # Minimum boundary is set set in sysconfig/sapconf
        declare -i ${name}_MIN=0
    done

    # Read current kernel tuning values from sysctl
    declare -i SHMMAX=$(sysctl -n kernel.shmmax)
    declare -i SHMALL=$(sysctl -n kernel.shmall)
    declare -i MAX_MAP_COUNT=$(sysctl -n vm.max_map_count)
    read -r SEMMSL SEMMNS SEMOPM SEMMNI < <(sysctl -n kernel.sem)

    # Read current tmpfs mount options and size
    read discard discard discard TMPFS_OPTS discard < <(grep -E '^tmpfs /dev/shm .+' /proc/mounts)
    if [ ! "$TMPFS_OPTS" ]; then
        echo "The system does not use tmpfs. Please configure tmpfs and try again." 1>&2
        exit 1
    fi
    # Remove size= from mount options
    TMPFS_OPTS=$(echo "$TMPFS_OPTS" | sed 's/size=[^,]\+//' | sed 's/^,//' | sed 's/,$//' | sed 's/,,/,/')
    declare -i TMPFS_SIZE=$(($(stat -fc '(%b*%S)>>10' /dev/shm))) # in KBytes

    # Read minimal value requirements from sysconfig
    if [ -r /etc/sysconfig/sapconf ]; then
        source /etc/sysconfig/sapconf
    fi
    
    # Calculate tuning parameter recommendations according to SAP notes
    declare -i SHMALL_REQ=$((VSZ*1024/PSZ)) # Note 941735: kernel.shmall is in pages; 20GB
    declare -i SHMMAX_REQ=$((VSZ*1024)) # Note 941735:  kernel.shmmax is in Bytes: 20GB
    declare -i TMPFS_SIZE_REQ=$(((VSZ*VSZ_TMPFS_PERCENT)/100)) # Note 941735: size of tmpfs in KB (RAM + SWAP) * 0.75
    echo $VSZ $VSZ_TMPFS_PERCENT $TMPFS_SIZE_REQ

    # No value may go below minimal
    for name in $TUNED_VARS; do
        min=${name}_MIN
        req=${name}_REQ
        val=${!name}
        if [ ! "${!req}" -o "${!req}" -lt "$val" ]; then
            declare -i $req="$val"
        fi
        if [ "${!req}" -lt "${!min}" ]; then
            declare -i $req="${!min}"
        fi
    done

    # Tune tmpfs - enlarge if necessary
    save_value tmpfs.size "$TMPFS_SIZE"
    save_value tmpfs.mount_opts "$TMPFS_OPTS"
    if [ "$TMPFS_SIZE_REQ" -gt "$TMPFS_SIZE" ]; then
        mount -o "remount,${TMPFS_OPTS},size=${TMPFS_SIZE_REQ}k" /dev/shm
    fi

    # Tune kernel parameters
    save_value kernel.shmmax "$SHMMAX"
    save_value kernel.sem "$(sysctl -n kernel.sem)"
    save_value kernel.shmall "$SHMALL"
    save_value vm.max_map_count "$MAX_MAP_COUNT"
    sysctl -w kernel.shmmax="$SHMMAX_REQ"
    sysctl -w kernel.sem="$SEMMSL_REQ $SEMMNS_REQ $SEMOPM_REQ $SEMMNI_REQ"
    sysctl -w kernel.shmall="$SHMALL_REQ"
    sysctl -w vm.max_map_count="$MAX_MAP_COUNT_REQ"

    # Tune ulimits for the max number of open files (rollback is not necessary in stop function)
    all_nofile_limits=""
    for limit in ${!LIMIT_*}; do # LIMIT_ parameters originate from sysconf/sapconf
        all_nofile_limits="$all_nofile_limits\n${!limit}"
    done
    for ulimit_group in @sapsys @sdba @dba; do
        for ulimit_type in soft hard; do
            sysconf_line=$(echo -e "$all_nofile_limits" | grep -E "^${ulimit_group}[[:space:]]+${ulimit_type}[[:space:]]+nofile.+")
            limits_line=$(grep -E "^${ulimit_group}[[:space:]]+${ulimit_type}[[:space:]]+nofile.+" /etc/security/limits.conf)
            if [ "$limits_line" ]; then
                sed -i "/$limits_line/d" /etc/security/limits.conf
            fi
            echo "$sysconf_line" >> /etc/security/limits.conf
        done
    done

    # SAP note 1557506 - Linux paging improvements
    # Tune pagecache_limit_mb and set pagecache_limit_ignore_dirty = 2
    # Note that the calculation is based on system RAM size instead of VSZ.
    declare -ri MEMSIZE_GB=$(($(grep MemTotal /proc/meminfo | awk '{print $2}')/1024/1024))
    declare -ri PAGECACHE_LIMIT=$(sysctl -n vm.pagecache_limit_mb)
    if [ "$MEMSIZE_GB" -lt 16 ]; then
        declare -i PAGECACHE_LIMIT_NEW=512
    elif [ "$MEMSIZE_GB" -lt 32 ]; then
        declare -i PAGECACHE_LIMIT_NEW=1024
    elif [ "$MEMSIZE_GB" -lt 64 ]; then
        declare -i PAGECACHE_LIMIT_NEW=2048
    else
        declare -i PAGECACHE_LIMIT_NEW=4096
    fi
    if [ "$PAGECACHE_LIMIT_NEW" -lt "$PAGECACHE_LIMIT" ]; then
        declare -i PAGECACHE_LIMIT_NEW="$PAGECACHE_LIMIT"
    fi

    # Apply new pagecache settings
    save_value vm.pagecache_limit_mb "$PAGECACHE_LIMIT"
    sysctl -w "vm.pagecache_limit_mb=$PAGECACHE_LIMIT_NEW"
    save_value vm.pagecache_limit_ignore_dirty $(sysctl -n vm.pagecache_limit_ignore_dirty)
    sysctl -w "vm.pagecache_limit_ignore_dirty=2"

    return 0
}

stop() {
    # Restore tuned kernel parameters
    SHMMAX=$(restore_value kernel.shmmax)
    SEM=$(restore_value kernel.sem)
    SHMALL=$(restore_value kernel.shmall)
    MAX_MAP_COUNT=$(restore_value vm.max_map_count)
    [ "$SHMMAX" ] && sysctl -w kernel.shmmax="$SHMMAX"
    [ "$SHMALL" ] && sysctl -w kernel.shmall="$SHMALL"
    [ "$SEM" ] && sysctl -w kernel.sem="$SEM"
    [ "$MAX_MAP_COUNT" ] && sysctl -w vm.max_map_count="$MAX_MAP_COUNT"

    # Restore the size of tmpfs
    TMPFS_SIZE=$(restore_value tmpfs.size)
    TMPFS_OPTS=$(restore_value tmpfs.mount_opts)
    [ "$TMPFS_SIZE" -a -e /dev/shm ] && mount -o "remount,${TMPFS_OPTS},size=${TMPFS_SIZE}k" /dev/shm

    # Restore pagecahce settings
    PAGECACHE_LIMIT=$(restore_value vm.pagecache_limit_mb)
    [ "$PAGECACHE_LIMIT" ] && sysctl -w "vm.pagecache_limit_mb=$PAGECACHE_LIMIT"
    PAGECACHE_LIMIT_IGNORE_DIRTY=$(restore_value vm.pagecache_limit_ignore_dirty)
    [ "$PAGECACHE_LIMIT_IGNORE_DIRTY" ] && sysctl -w "vm.pagecache_limit_ignore_dirty=$PAGECACHE_LIMIT_IGNORE_DIRTY"

    return 0
}

process $@
